color_code_black_on_white = f"\033[0m"
color_code_white_on_red = f"\033[97;41m"
color_code_white_on_black = f"\033[97;40m"
color_code_white_on_blue = f"\033[97;44m"

color_code_black_on_white_v2 = "\033[1;1H\033[2J"
